#ifndef __CBOARD__
#define __CBOARD__

#include "bitboard.h"

#include "debug.h"

#include "types.h"

typedef enum {
	A1 = 0, B1, C1, D1, E1, F1, G1, H1,
	A2, B2, C2, D2, E2, F2, G2, H2,
	A3, B3, C3, D3, E3, F3, G3, H3,
	A4, B4, C4, D4, E4, F4, G4, H4,
	A5, B5, C5, D5, E5, F5, G5, H5,
	A6, B6, C6, D6, E6, F6, G6, H6,
	A7, B7, C7, D7, E7, F7, G7, H7,
	A8, B8, C8, D8, E8, F8, G8, H8,
	BAD_SQUARE
} squares;

typedef enum {FILEA = 0, FILEB, FILEC, FILED, FILEE, FILEF, FILEG, FILEH} files;
typedef enum {RANK1 = 0, RANK2, RANK3, RANK4, RANK5, RANK6, RANK7, RANK8} ranks;

typedef enum {FREE = 0, PAWN = 1, KNIGHT, BISHOP, ROOK, QUEEN, KING} pieces;
//typedef enum {PAWN_V = 1000, KNIGHT_V = 2600, BISHOP_V = 2900, ROOK_V = 4500, QUEEN_V = 8800, KING_V = 99000} values;
typedef enum {PAWN_V = 1000, KNIGHT_V = 3000, BISHOP_V = 3000, ROOK_V = 5000, QUEEN_V = 9000, KING_V = 99000} values;

const int piece_v[7] = {0, PAWN_V, KNIGHT_V, BISHOP_V, ROOK_V, QUEEN_V, KING_V};

// side to move
const int WHITE =  1;
const int BLACK =  0;
				   
// castle flags
const int WHITE_LCASTLE = 0x1; // long castling for white
const int WHITE_SCASTLE = 0x2; // short castling for white
const int BLACK_LCASTLE = 0x4; // long castling for black
const int BLACK_SCASTLE = 0x8; // short castling for black

const int WHITE_HAS_CASTLED = 0x10;
const int BLACK_HAS_CASTLED = 0x20;

#define WhiteCastled() (board.castle_flags & WHITE_HAS_CASTLED)
#define BlackCastled() (board.castle_flags & BLACK_HAS_CASTLED)

class Board {
public:
	
	// minimal board info
	BITBOARD 
		wPawns,
		wRooks,
		wKnights,
		wBishops,
		wQueens,
		wKing,
		bPawns,
		bRooks,
		bKnights,
		bBishops,
		bQueens,
		bKing;

	int ep_square;	// 0 if last move wasn't ep
					// else the intermediarry square

	int side_to_move;

	int castle_flags;

	char age;

	// extra information
	BITBOARD all;		// or of all pieces
	BITBOARD rotated_90,	// all rotated 90 degress
			 diag_a1, // rotated diag a1-h8
			 diag_h1; // rotated diag a8-h1
	BITBOARD white_pieces, black_pieces;

	BITBOARD rooks_queens, bishops_queens;

	int white_king_sq, black_king_sq;

	int sq[64]; // see pieces enum above, black pieces have negative values

	HASHCODE hashcode;
	HASHCODE pawn_hashcode;

	int white_material, white_pawns; // number of white pieces and white pawns
	int black_material, black_pawns; // number of black pieces and black pawns
	int material;					 // total material value

	int piecetable_score;

	int piece_counter[2][6]; // counter for 0=UNUSED/PAWN/KNIGHT/BISHOP/ROOK/QUEEN

	void EvalPiecetable();

	void ComputeMaterial();
	void VerifyMaterial();

	HASHCODE ComputeHashcode(int wtm);
	void StoreHashcode(int wtm);
	void UpdateHash(int move, int wtm);

	HASHCODE ComputePawnHashcode(int wtm);
	void StorePawnHashcode(int wtm);

	void Init(char* str);
	void Print();
	char* PrintBuf(); // prints the board to a buffer
	char* PrintEPD();

	int* GenerateAllMoves(int* move, int wtm);
	int* GenerateAllValidMoves(int* move, int wtm);

	int* GenerateCaptures(int* move, int wtm);
	int* GenerateNonCaptures(int* move, int wtm);
	int* GenerateCheckEvasions(int* move, int wtm);
	
	void Make(int move, int ply, int wtm);
	void Unmake(int move, int ply, int wtm);

	void Verify(char* where="Unknown");
	void VerifyHash(int wtm);
	void VerifyPawnHash(int wtm);
	void PrintDebug(); // internal use only

	int Evaluate(int alpha, int beta, int ply, int wtm);

	int Analyse(int wtm, int& next_possible_move, int max_ply = 60, int best_move = 0);
	void AnalyseEPD(char* filename, char* outfile, int maxtime, int maxply, char* descr = "unknown configuration");

	void AnalyseAllMoves(int wtm, int max_ply = 60, int best_move = 0);
	void AnalyseAllMovesEPD(char* filename, char* outfile, int maxtime, int maxply, char* descr = "unknown configuration");

	void Reset(); // set it to startup position

	int IsValidMove(int move, int wtm);

	void InitPieceCounters();

	void VerifyPieceCounter();

	void Mirror();

	void CheckMirroredEvalEPD(char* filename, char* outfile);
};

extern Board board;

// Board macros
#define PieceOnSquare(i) board.sq[i]

// Lookup Tables

typedef struct {	// a8h1 is the left diagonal, a1h8 the right one
	BITBOARD a1[256];
	int shift_a1;	
	BITBOARD h1[256];
	int shift_h1;	
	// no mask is necessary, as the rotated bitboards are so computed as the mask will 
	// always be 0xff
} DiagInfo;


extern BITBOARD rook_attacks[64][256];
extern BITBOARD rook_attacks90[64][256];

extern DiagInfo bishop_attacks[64];

extern BITBOARD knight_attacks[64];
extern BITBOARD king_attacks[64];

extern BITBOARD wpawn_attacks[64];
extern BITBOARD bpawn_attacks[64];

extern BITBOARD rank_mask[8];
extern BITBOARD file_mask[8]; 

extern BITBOARD set_mask[64]; 

extern BITBOARD mask_left_edge;
extern BITBOARD mask_right_edge;

extern BITBOARD clear_mask_90[64];
extern BITBOARD clear_mask_a1[64];
extern BITBOARD clear_mask_h1[64];

extern BITBOARD set_mask_90[64];
extern BITBOARD set_mask_a1[64];
extern BITBOARD set_mask_h1[64];

// structs for saving move information
struct SaveInfo {
	int ep_move;	// 1 if ep, 0 if not
};

#define MAX_PLY 1024

extern int move_list[];
extern int last_move[];	// pointer in move_list;

extern int current_move[]; // current move at each ply
extern int pv[][MAX_PLY];  // principal variation
extern int pvlen[];		   // length of each pv

extern BITBOARD white_scastle_mask, white_lcastle_mask, black_scastle_mask, black_lcastle_mask;

extern SaveInfo save_struct[MAX_PLY];


// useful macros
#define File(x) ((x)&7)
#define Rank(x) ((x)>>3)

#if 1

#define AttacksRank(a) rook_attacks[a][Shr(board.all, a & 0xf8) & 0xff]
#define AttacksFile(a) rook_attacks90[a][Shr(board.rotated_90, (a & 7) << 3) & 0xff]
#define AttacksDiagA1(a) bishop_attacks[a].a1[Shr(board.diag_a1, bishop_attacks[a].shift_a1) & 0xff]
#define AttacksDiagH1(a) bishop_attacks[a].h1[Shr(board.diag_h1, bishop_attacks[a].shift_h1) & 0xff]

#else 

BITBOARD __fastcall AttacksFile(int sq);
BITBOARD __fastcall AttacksRank(int sq);
BITBOARD __fastcall AttacksDiagA1(int sq);
BITBOARD __fastcall AttacksDiagH1(int sq);

#endif

#define AttacksBishop(a) Or(AttacksDiagA1(a), AttacksDiagH1(a))
#define AttacksRook(a) Or(AttacksRank(a), AttacksFile(a))
#define AttacksQueen(a) Or(AttacksRook(a), AttacksBishop(a))

#define Clear90(a) board.rotated_90 = And(board.rotated_90, clear_mask_90[a])
#define ClearA1(a) board.diag_a1 = And(board.diag_a1, clear_mask_a1[a]);
#define ClearH1(a) board.diag_h1 = And(board.diag_h1, clear_mask_h1[a]);

#define Set90(a) board.rotated_90 = Or(board.rotated_90, set_mask_90[a])
#define SetA1(a) board.diag_a1 = Or(board.diag_a1, set_mask_a1[a]);
#define SetH1(a) board.diag_h1 = Or(board.diag_h1, set_mask_h1[a]);

#define ClearBoardBit(a) Clear(board.all, a); Clear90(a); ClearA1(a); ClearH1(a);
#define SetBoardBit(a) Set(board.all, a); Set90(a); SetA1(a); SetH1(a);

#define SetClear(bitboard, from, to) Clear(bitboard, from); Set(bitboard,to);
#define SetClearBoardBit(from, to) ClearBoardBit(from); SetBoardBit(to);

#define NextSide(wtm) (1-(wtm))

#define FirstMove(ply) (move_list+last_move[ply-1]+1)

#define KingSquare(wtm) ((wtm) ? board.white_king_sq : board.black_king_sq)

#define Free(square) (sq[square] ? 0 : 1)

#define Check(wtm) Attacked(KingSquare(wtm), 1-wtm)

#define PLY_DEPTH 100

#define INF 327680

#define MATE (INF-10)
#define MIN_MATE_SCORE 320000

#define DrawScore(wtm) 0
#define MateScore(wtm) ((wtm)?(INF-1):(-INF+1))

#define Max(a,b) (((a)>(b))?(a):(b))
#define Min(a,b) (((a)<(b))?(a):(b))
#define Abs(a) ((a>=0)?(a):-(a))


// useful functions
BITBOARD AttacksFrom(int square, int wtm);
BITBOARD AttacksTo(int square, int wtm);
int Attacked(int square, int wtm);


#define Square(File, Rank) (((File)<<3)+(Rank))

#endif